# VERSION: MAJOR.MINOR.BUGFIX
# When new functionality is added update minor. For new builds with only
# bugfixes etc update bugfix number.

# Pypi needs a new version number for each publication.

__version__ = '1.0.3'